/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComPrediction.h
    \brief    prediction class (header)
*/

#ifndef __TCOMPREDICTION__
#define __TCOMPREDICTION__


// Include files
#include "TComPic.h"
#include "TComMotionInfo.h"
#include "TComPattern.h"
#include "TComTrQuant.h"
#include "TComInterpolationFilter.h"
#include "TComWeightPrediction.h"

#if SVC_UPSAMPLING
#include "TComDataCU.h"
#include "TComUpsampleFilter.h"
#endif
#if SHVC_COMPLEXITY_ASSESSMENT
#include "TComMemCalc.h"
#endif //SHVC_COMPLEXITY_ASSESSMENT
//! \ingroup TLibCommon
//! \{

// ====================================================================================================================
// Class definition
// ====================================================================================================================

/// prediction class
class TComPrediction : public TComWeightPrediction
{
protected:
#if FDP_COMBINED_INTER_MODE
	Pel* m_predFDP_COMBINED_INTER_MODE;
	Int * m_predFDP_COMBINED_INTER_MODEfilter;
#endif
  Int*      m_piYuvExt;
  Int       m_iYuvExtStride;
  Int       m_iYuvExtHeight;
  
  TComYuv   m_acYuvPred[2];
  TComYuv   m_cYuvPredTemp;
  TComYuv m_filteredBlock[4][4];
  TComYuv m_filteredBlockTmp[4];
  
  TComInterpolationFilter m_if;

#if SVC_UPSAMPLING
  TComUpsampleFilter      m_cUsf;
#endif
#if FDP_COMBINED_INTER_MODE
  TComYuv   m_acBLRec;
#endif
#if SHVC_COMPLEXITY_ASSESSMENT
  TComFrameMemoryAccessCalculator* m_apcMemAccessCalc[ NUM_MEMORY_ARCHITECTURES ];
  TComFrameMemoryAccessCalculator* m_apcMemAccessCalcBlock[ NUM_MEMORY_ARCHITECTURES ];
  TComFrameMemoryAccessCalculator* m_apcMemAccessCalcPic[ NUM_MEMORY_ARCHITECTURES ];
#endif // SHVC_COMPLEXITY_ASSESSMENT
  Pel*   m_pLumaRecBuffer;       ///< array for downsampled reconstructed luma sample 
  Int    m_iLumaRecStride;       ///< stride of #m_pLumaRecBuffer array
  UInt   m_uiaShift[ 63 ];       // Table for multiplication to substitue of division operation

  Void xPredIntraAng            ( Int* pSrc, Int srcStride, Pel*& rpDst, Int dstStride, UInt width, UInt height, UInt dirMode, Bool blkAboveAvailable, Bool blkLeftAvailable, Bool bFilter );
  Void xPredIntraPlanar         ( Int* pSrc, Int srcStride, Pel* rpDst, Int dstStride, UInt width, UInt height );
  
  // motion compensation functions
  Void xPredInterUni            ( TComDataCU* pcCU,                          UInt uiPartAddr,               Int iWidth, Int iHeight, RefPicList eRefPicList, TComYuv*& rpcYuvPred, Int iPartIdx, Bool bi=false          );
#if FDP_COMBINED_INTER_MBW_HEVC
  Void xPredInterBi             ( TComDataCU* pcCU,                          UInt uiPartAddr,               Int iWidth, Int iHeight,                         TComYuv*& rpcYuvPred, Int iPartIdx, Bool& biPred           );
#else
  Void xPredInterBi             ( TComDataCU* pcCU,                          UInt uiPartAddr,               Int iWidth, Int iHeight,                         TComYuv*& rpcYuvPred, Int iPartIdx          );
#endif
  Void xPredInterLumaBlk  ( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi );
  Void xPredInterChromaBlk( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi );
  Void xWeightedAverage         ( TComDataCU* pcCU, TComYuv* pcYuvSrc0, TComYuv* pcYuvSrc1, Int iRefIdx0, Int iRefIdx1, UInt uiPartAddr, Int iWidth, Int iHeight, TComYuv*& rpcYuvDst );
  
  Void xGetLLSPrediction ( TComPattern* pcPattern, Int* pSrc0, Int iSrcStride, Pel* pDst0, Int iDstStride, UInt uiWidth, UInt uiHeight, UInt uiExt0 );

  Void xDCPredFiltering( Int* pSrc, Int iSrcStride, Pel*& rpDst, Int iDstStride, Int iWidth, Int iHeight );
  Bool xCheckIdenticalMotion    ( TComDataCU* pcCU, UInt PartAddr);

public:
  TComPrediction();
  virtual ~TComPrediction();
  
  Void    initTempBuff();
#if SHVC_COMPLEXITY_ASSESSMENT
  Void initMCMemoryAccessCalculator ( TComPic* pcRefPic, int design_flag);
  Void xCalcMCMemoryAccessLuma      ( TComDataCU* pcCU, Int uiPartAddr, RefPicList eRefPicList, Int iRefIdx, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV);
  Void xCalcMCMemoryAccessChroma    ( TComDataCU* pcCU, Int uiPartAddr, RefPicList eRefPicList, Int iRefIdx, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV);
  Void xCalcUSMemoryAccessLuma      ( TComDataCU* pcCU, Int uiPartAddr, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV);
  Void xCalcUSMemoryAccessChroma    ( TComDataCU* pcCU, Int uiPartAddr, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV);
#endif //SHVC_COMPLEXITY_ASSESSMENT  
#if INTRA_BL
  Void getBaseBlk ( TComDataCU* pcCU, TComYuv* pcYuvPred, Int iPartAddr, Int iWidth, Int iHeight );
#endif
  // inter
  Void motionCompensation         ( TComDataCU*  pcCU, TComYuv* pcYuvPred, RefPicList eRefPicList = REF_PIC_LIST_X, Int iPartIdx = -1 );
  
  // motion vector prediction
  Void getMvPredAMVP              ( TComDataCU* pcCU, UInt uiPartIdx, UInt uiPartAddr, RefPicList eRefPicList, Int iRefIdx, TComMv& rcMvPred );
  
  // Angular Intra
  Void predIntraLumaAng           ( TComPattern* pcTComPattern, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight,  TComDataCU* pcCU, Bool bAbove, Bool bLeft );
  Void predIntraChromaAng         ( TComPattern* pcTComPattern, Int* piSrc, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, TComDataCU* pcCU, Bool bAbove, Bool bLeft );
  
  Pel  predIntraGetPredValDC      ( Int* pSrc, Int iSrcStride, UInt iWidth, UInt iHeight, Bool bAbove, Bool bLeft );
  
  Int* getPredicBuf()             { return m_piYuvExt;      }
  Int  getPredicBufWidth()        { return m_iYuvExtStride; }
  Int  getPredicBufHeight()       { return m_iYuvExtHeight; }

#if !REMOVE_LMCHROMA
  Void predLMIntraChroma( TComPattern* pcPattern, Int* piSrc, Pel* pPred, UInt uiPredStride, UInt uiCWidth, UInt uiCHeight, UInt uiChromaId );
  Void getLumaRecPixels  ( TComPattern* pcPattern, UInt uiWidth0, UInt uiHeight0 );
#endif

#if SVC_UPSAMPLING
  Void upsampleBasePic( TComPicYuv* pcUsPic, TComPicYuv* pcBasePic, TComPicYuv* pcTempPic );
#endif
#if FDP_COMBINED_INTER_MODE
#if FDP_COMBINED_INTER_MBW_HEVC
  Void combinedPrediction( TComDataCU* pcCU, Pel* pDst, UInt uiDstStride, UInt uiAbsPartIdx, Int iWidth, Int iHeight, Bool biPred =false);
#else
  Void combinedPrediction( TComDataCU* pcCU, Pel* pDst, UInt uiDstStride, UInt uiAbsPartIdx, Int iWidth, Int iHeight );
#endif
#if FDP_COMBINED_INTER_MODE	   
  Void 	filtersmoothFDP_COMBINED_INTER_MODEhor(Pel *A,Int *B,Int limit);
  Void  filtersmoothFDP_COMBINED_INTER_MODEver(Int *A, Pel*B,Int limit,Int step);
#endif
#endif
#if SHVC_COMPLEXITY_ASSESSMENT
  Void   initMCMemoryAccessCalculator( const MemCmpParam& cLumaParam, const MemCmpParam& cChromaParam, int design_flag );
  Void   updatePic( int design_flag );
  UInt64 getTotalMCMemoryAccessBytes( Int iIndex, int design_flag );
#endif //SHVC_COMPLEXITY_ASSESSMENT
};

//! \}

#endif // __TCOMPREDICTION__
